"""Given a PubMed ID (PMID), return a list of publications which cite it"""
# https://icite.od.nih.gov/api

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from sys import stdout
from os.path import exists
from os.path import join
import collections as cx

from pmidcite.cli.utils import read_top_pmids
from pmidcite.icite.entry import NIHiCiteEntry
from pmidcite.icite.paper import NIHiCitePaper
from pmidcite.icite.pmid_loader import NIHiCiteLoader


class NIHiCiteDownloader:
    """Manage pubs notes files"""

    def __init__(self, force_dnld, api, rpt_references=False):
        self.rpt_references = rpt_references
        self.dnld_force = force_dnld
        self.dir_dnld = api.dir_dnld  # e.g., ./icite
        self.nihgrouper = api.nihgrouper
        self.loader = NIHiCiteLoader(self.nihgrouper, self.dir_dnld, prt=None)
        self.api = api                # NIHiCiteAPI

    def wr_papers(self, fout_txt, pmid2icitepaper, force_overwrite=False, mode='w'):
        """Run iCite for user-provided PMIDs and write to a file"""
        if not pmid2icitepaper:
            return
        pmids_all = pmid2icitepaper.keys()
        pmids_new = pmids_all
        if mode == 'a':
            pmids_new = self._get_new_pmids(fout_txt, pmids_all)
        if pmids_new:
            if self._do_write(fout_txt, force_overwrite):
                with open(fout_txt, mode) as prt:
                    self.prt_papers(pmid2icitepaper, prt)
                print('{WR}: {TXT}'.format(
                    WR=self._msg_wrote(mode, pmids_all, pmids_new), TXT=fout_txt))

    @staticmethod
    def prt_hdr(prt=stdout):
        """Print column headers in one line"""
        prt.write('TYP {HDR}\n'.format(HDR=NIHiCiteEntry.hdr))

    @staticmethod
    def prt_keys(prt=stdout):
        """Print paper keys"""
        prt.write('\nKEYS TO PAPER LINE:\n')
        prt.write('    TYP {ICITE_FMT}\n'.format(ICITE_FMT=NIHiCiteEntry.line_fmt()))
        prt.write('\n')
        prt.write('TYPe of relationship to the user-requested paper (TYP):\n')
        NIHiCitePaper.prt_keys(prt)
        prt.write('\nNIH iCite details:\n\n')
        NIHiCiteEntry.prt_key_desc(prt)
        prt.write('\n')

    def prt_papers(self, pmid2icitepaper, prt=stdout, prt_assc_pmids=True):
        """Print papers, including citation counts, cite_by and references list"""
        for pmid, paper in pmid2icitepaper.items():
            if paper is not None:
                self.prt_paper(paper, pmid, pmid, prt, prt_assc_pmids)
            else:
                print('**WARNING: NO iCite ENTRY FOUND FOR: {PMID}'.format(PMID=pmid))

    # pylint: disable=too-many-arguments
    def prt_paper(self, paper, pmid, name, prt=stdout, prt_assc_pmids=True):
        """Print one paper, including citation counts, cite_by and references list"""
        if paper is not None:
            if prt_assc_pmids:
                paper.prt_summary(prt, self.rpt_references,
                                  sortby_cites='nih_sd',
                                  sortby_refs='nih_sd')
                prt.write('\n')
            else:
                prt.write('TOP {iCite}\n'.format(iCite=paper.str_line()))
        else:
            prt.write('No iCite results found: {PMID} {NAME}\n\n'.format(
                PMID=pmid, NAME=name if name is not None else ''))

    @staticmethod
    def _msg_wrote(mode, pmids_req, pmids_new):
        """Get the 'WROTE' or 'APPENDED' message"""
        if mode == 'w':
            return '{N:6,} WROTE'.format(N=len(pmids_req))
        if mode == 'a':
            if pmids_new:
                return '{N:,} of {M:,} APPENDED'.format(
                    N=len(pmids_new),
                    M=len(pmids_req))
            return '{N:,} of {M:,} FOUND'.format(
                N=len(pmids_new),
                M=len(pmids_req))
            # return '{N:,} FOUND'.format(M=len(pmids_req))
        raise RuntimeError('UNRECOGNIZED WRITE MODE({M})'.format(M=mode))

    @staticmethod
    def _get_new_pmids(pmidcite_txt, pmids):
        """Get PMIDs which are not already fully analyzed in pmidcite.txt"""
        if not exists(pmidcite_txt):
            return pmids
        pmids_old = read_top_pmids(pmidcite_txt)
        return [p for p in pmids if p not in pmids_old]

    def wr_name2pmid(self, fout_txt, name2pmid):
        """Run iCite for user-provided PMIDs and write to a file"""
        # pylint: disable=line-too-long
        name2ntpaper = self._run_icite_name2pmid(name2pmid, dnld_assc_pmids_do=False, pmid2note=None)
        if name2ntpaper:
            with open(fout_txt, 'w') as prt:
                for name, ntpaper in name2ntpaper.items():
                    self.prt_paper(ntpaper.paper, ntpaper.pmid, name, prt)
                print('  WROTE: {TXT}'.format(TXT=fout_txt))

    def _run_icite_name2pmid(self, name2pmid, dnld_assc_pmids_do, pmid2note):
        """Get a NIHiCitePaper object for each user-specified PMID"""
        name2ntpaper = {}
        ntobj = cx.namedtuple('Paper', 'pmid paper')
        for name, pmid in name2pmid.items():
            paper = self._geticitepaper(pmid, name, dnld_assc_pmids_do, pmid2note)
            name2ntpaper[name] = ntobj(pmid=pmid, paper=paper)
        return name2ntpaper

    def get_pmid2paper(self, pmids, dnld_assc_pmids_do=True, pmid2note=None):
        """Get a NIHiCitePaper object for each user-specified PMID"""
        s_geticitepaper = self._geticitepaper
        # Arg, dnld_assc_pmids_do, causes iCite data to be downloaded if needed, or loaded
        if not pmid2note:
            papers = [s_geticitepaper(p, '', dnld_assc_pmids_do, None) for p in pmids]
        else:
            papers = [s_geticitepaper(p, '', dnld_assc_pmids_do, pmid2note) for p in pmids]
        # Note: if there is no iCite entry for a PMID, paper will be None
        return cx.OrderedDict(zip(pmids, papers))  # pmid2ntpaper

    def _geticitepaper(self, pmid_top, header, dnld_assc_pmids_do, pmid2note):
        """Print summary for each user-specified PMID"""
        citeobj_top = self.get_icite(pmid_top)  # NIHiCiteEntry
        if citeobj_top:
            if dnld_assc_pmids_do:
                self._dnld_assc_pmids(citeobj_top)
            icites = self.loader.load_icite_mods_all([pmid_top])
            pmid2icite = {o.dct['pmid']:o for o in icites}
            return NIHiCitePaper(pmid_top, pmid2icite, header, pmid2note)
        note = ''
        if pmid2note and pmid_top in pmid2note:
            note = pmid2note[pmid_top]
        print('No NIH iCite results found: {PMID} {HDR} {NOTE}'.format(
            PMID=pmid_top,
            HDR=header if header else '',
            NOTE=note))
        return None  ## TBD: NIHiCitePaper(pmid_top, self.dir_dnld, header, note)

    def _init_iciteobj(self, icite_dct):
        """Initialize the top NIH iCite paper, if it exists"""
        if icite_dct:
            return NIHiCiteEntry(icite_dct, self.nihgrouper.get_group(icite_dct['nih_percentile']))
        return None

    def _dnld_assc_pmids(self, icite):
        """Download PMID iCite data for PMIDs associated with icite paper"""
        pmids_assc = icite.get_assc_pmids()
        if not pmids_assc:
            return []
        if self.dnld_force:
            return self.api.dnld_icites(pmids_assc)
        pmids_missing = self._get_pmids_missing(pmids_assc)
        if pmids_missing:
            objs_missing = self.api.dnld_icites(pmids_missing)
            pmids_load = pmids_assc.difference(pmids_missing)
            objs_dnlded = self.loader.load_icites(pmids_load)
            return objs_missing + objs_dnlded
        return self.loader.load_icites(pmids_assc)

    def _get_pmids_missing(self, pmids_all):
        """Get PMIDs that have not yet been downloaded"""
        pmids_missing = set()
        for pmid_cur in pmids_all:
            file_pmid = join(self.dir_dnld, 'p{PMID}.py'.format(PMID=pmid_cur))
            if not exists(file_pmid):
                pmids_missing.add(pmid_cur)
        return pmids_missing

    def get_icites(self, pmids):
        """Load or download NIH iCite data for requested PMIDs"""
        icites = []
        for pmid in pmids:
            icite = self.get_icite(pmid)
            if icite is not None:
                icites.append(icite)
        return icites

    def get_icite(self, pmid):
        """Load or download NIH iCite data for requested PMID"""
        file_pmid = join(self.dir_dnld, 'p{PMID}.py'.format(PMID=pmid))
        if self.dnld_force or not exists(file_pmid):
            iciteobj = self.api.dnld_icite(pmid)
            if iciteobj is not None:
                return iciteobj
        return self.loader.load_icite(file_pmid)  # NIHiCiteEntry

    @staticmethod
    def _do_write(fout_txt, force_overwrite):
        """Ask for a yes-no answer from the user on STDIN"""
        if not exists(fout_txt) or force_overwrite:
            return True
        prompt_user = '\nover-write {TXT} (yes/no)? '.format(TXT=fout_txt)
        return input(prompt_user).lower()[:1] == 'y'


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Run PubMed user query and download PMIDs. Run iCite on PMIDs. Write text file."""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import os
import sys
from collections import namedtuple
from pmidcite.cfg import get_cfgparser
from pmidcite.eutils.cmds.pubmed import PubMed
from pmidcite.cli.utils import wr_pmids
from pmidcite.icite.run import PmidCite


class PubMedQueryToICite:
    """Run PubMed user query and download PMIDs. Run iCite on PMIDs. Write text file."""

    def __init__(self, force_dnld, verbose=True, pmid2note=None, prt_icitepy=None):
        self.force_dnld = force_dnld
        self.verbose = verbose
        self.pmid2note = {} if pmid2note is None else pmid2note
        # Setting prt_icitepy to sys.stdout causes messages: WROTE: ./icite/p31898878.py
        self.prt_icitepy = prt_icitepy
        self.pmidcite = PmidCite(get_cfgparser())
        cfg = self.pmidcite.cfgparser
        self.pubmed = PubMed(
            email=cfg.get_email(),
            apikey=cfg.get_apikey(),
            tool=cfg.get_tool())
        self._chk_dirs()

    def run(self, lst, dnld_idxs=None):
        """Give an list of tuples: Query PubMed for PMIDs. Download iCite, given PMIDs"""
        # Download PMIDs and NIH's iCite for only one PubMed query
        nts = self.get_nts_g_list(lst)
        if dnld_idxs is not None:
            for idx in dnld_idxs:
                ntd = nts[idx]
                self.querypubmed_runicite(ntd.filename, ntd.pubmed_query)
        # Download PMIDs and NIH's iCite for all PubMed queries
        else:
            for ntd in nts:
                ## print('QQQQQQQQQQQQQQQQ pmidcite/pubmedqueryicite.py', ntd.pubmed_query)
                self.querypubmed_runicite(ntd.filename, ntd.pubmed_query)

    def querypubmed_runicite(self, fout_pat, query):
        """Given a user query, return PMIDs. Then run NIH's iCite"""
        # 1) Query PubMed and download PMIDs
        pmids = self.pubmed.dnld_query_pmids(query)
        fout_pmids = os.path.join(self.get_dir_pmids(), fout_pat.format(PRE='pmids'))
        fout_icite = os.path.join(self.get_dir_icite(), fout_pat.format(PRE='icite'))
        # 2) Write PubMed PMIDs into a simple text file, one PMID per line
        if fout_pmids != fout_icite:
            if pmids:
                wr_pmids(fout_pmids, pmids)
            else:
                print('  0 PMIDs: NOT WRITING {TXT}'.format(TXT=fout_pmids))
        # 3) Run NIH's iCite on the PMIDs and write the results into a file
        if pmids:
            self.wr_icite(fout_icite, pmids)

    def get_dir_pmids(self):
        """Get directory to store lists of PMIDs"""
        return self.pmidcite.cfgparser.cfgparser['pmidcite']['dir_pmids']

    def get_dir_icite(self):
        """Get directory to store lists of PMIDs"""
        return self.pmidcite.cfgparser.cfgparser['pmidcite']['dir_icite']

    def wr_icite(self, fout_icite, pmids, grouperobj=None):
        """Run PMIDs in iCite and print results into a file"""
        # Get NIHiCiteDownloader object
        dnldr = self.pmidcite.get_icitedownloader(
            self.force_dnld,
            grouperobj,
            no_references=False,
            prt_icitepy=self.prt_icitepy)
        pmid2paper = dnldr.get_pmid2paper(pmids, self.verbose, self.pmid2note)
        dnldr.wr_papers(fout_icite, pmid2icitepaper=pmid2paper, force_overwrite=True)

    @staticmethod
    def get_nts_g_list(lst):
        """Turn a list iof tuple strings into a list of namedtuples"""
        nto = namedtuple('Nt', 'filename pubmed_query')
        return [nto._make(t) for t in lst]

    @staticmethod
    def get_index(argv):
        """Get the index of the pubmed query to run"""
        # If no argument was provided, run the last query in the list
        if len(argv) == 1:
            return [-1]
        if argv[1] == 'all':
            return None
        return [int(n) for n in argv[1:] if n.lstrip('-').isdigit()]

    def _chk_dirs(self):
        """Check output directories for existance"""
        not_exist = set()
        dct = self.pmidcite.cfgparser.cfgparser['pmidcite']
        if not os.path.exists(dct['dir_pmids']):
            not_exist.add(dct['dir_pmids'])
        if not os.path.exists(dct['dir_icite']):
            not_exist.add(dct['dir_icite'])
        if not_exist:
            for dirname in not_exist:
                print('**FATAL: NO DIR: {DIR}'.format(DIR=dirname))
            sys.exit(1)


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



#!/usr/bin/env python

import matplotlib as mpl
mpl.use('agg')
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib import colors
import numpy as np
import scipy.stats as stats
import math

MU = 0
VARIANCE = 1
SIGMA = math.sqrt(VARIANCE)

def main():
    """Plot guassian distribution, showing icite groups"""
    fout_imgs = [
        '../pmidcite/doc/images/nih_perc_groups.png',
        '../bibliometrics/doc/nih_perc_groups.tiff',
        '../bibliometrics/doc/nih_perc_groups.pdf',
        '../bibliometrics/doc/nih_perc_groups.pdf',
    ]

    vlines = [-3, -2, -1, 1, 2, 3]
    xvals = np.linspace(MU - 3*SIGMA, MU + 3*SIGMA, 120)
    yvals = stats.norm.pdf(xvals, MU, SIGMA)
    x2y = {xvals:stats.norm.pdf(xvals, MU, SIGMA) for xvals in vlines}

    fig, axes = plt.subplots(1, 1)
    axes.plot(xvals, yvals, color='k')
    for xval in vlines:
        plt.vlines(xval, ymin=0, ymax=x2y[xval])
    plt.hlines(0, xmin=-3, xmax=3)

    # Fill with color
    colors = _get_colors('gist_rainbow')
    _fill_between(axes, -3, -2, colors[0])
    _fill_between(axes, -2, -1, colors[1])
    _fill_between(axes, -1, 1, colors[2])
    _fill_between(axes, 1, 2, colors[3])
    _fill_between(axes, 2, 3, colors[4])

    # Add text
    # 0) Lowest
    plt.text(-2.01, -.003, '0', fontsize=22, ha='right', va='bottom', fontweight='bold')
    # 1) Low
    kwtxt = {'ha':'center', 'va':'center', 'fontweight':'bold'}
    plt.text(-1.24, .12, '1', fontsize=25, **kwtxt)
    plt.text(-1.5, .03, 'Low\n14%', fontsize=15, **kwtxt)
    # 2) Good
    plt.text(0, .35, '2', fontsize=25, **kwtxt)
    plt.text(0, .03, 'Good\n68%', fontsize=15, **kwtxt)
    # 3) Better
    plt.text(1.24, .12, '3', fontsize=25, **kwtxt)
    plt.text(1.5, .03, 'High\n14%', fontsize=15, **kwtxt)
    # 4) Best
    plt.text(2.01, -.003, '4', fontsize=22, ha='left', va='bottom', fontweight='bold')

    plt.xlabel('Standard Deviation lines')
    plt.ylabel('Distribution')

    axes.annotate('2%', fontsize=15, xy=(2.5, 0.02),
            xycoords='data', xytext=(2.7, .07),
            arrowprops=dict(arrowstyle="->",
                            color = 'black'), **kwtxt)
    axes.annotate('2%', fontsize=15, xy=(-2.5, 0.02),
            xycoords='data', xytext=(-2.7, .07),
            arrowprops=dict(arrowstyle="->",
                            color = 'black'), **kwtxt)

    for fout_img in fout_imgs:
        plt.savefig(fout_img, dpi=800)
        print('WROTE: {IMG}'.format(IMG=fout_img))

def _get_colors(colormap):
    """Get colors"""
    cmap = cm.get_cmap(colormap)
    norm = colors.Normalize(vmin=0, vmax=4)
    scalarmap = cm.ScalarMappable(norm=norm, cmap=cmap)
    return [scalarmap.to_rgba(n) for n in range(5)]

def _fill_between(axes, xval0, xval1, facecolor):
    """Color between standard deviation lines"""
    xvals = np.linspace(xval0, xval1, 20)
    yvals = stats.norm.pdf(xvals, MU, SIGMA)
    axes.fill_between(xvals, yvals, facecolor=facecolor, alpha=.4)


if __name__ == '__main__':
    main()



"""Manage args for NIH iCite run for one PubMed ID (PMID)"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import os
from sys import stdout
import argparse

from pmidcite.eutils.cmds.pubmed import PubMed
from pmidcite.cfgini import prt_rcfile
from pmidcite.cli.utils import get_outfile
from pmidcite.cli.utils import get_pmids
from pmidcite.icite.nih_grouper import NihGrouper
from pmidcite.icite.pmid_dnlder import NIHiCiteDownloader


class NIHiCiteCli:
    """Manage args for NIH iCite run for one PubMed ID (PMID)"""

    def __init__(self, pmidcite):
        self.pmidcite = pmidcite
        cfgparser = self.pmidcite.cfgparser
        self.pubmed = PubMed(
            email=cfgparser.get_email(),
            apikey=cfgparser.get_apikey(),
            tool=cfgparser.get_tool())

    def get_argparser(self):
        """Argument parser for Python wrapper of NIH's iCite given PubMed IDs"""
        parser = argparse.ArgumentParser(description="Run NIH's iCite given PubMed IDs")
        dir_icite_py = self.pmidcite.cfgparser.cfgparser['pmidcite']['dir_icite_py']
        dir_icite = self.pmidcite.cfgparser.cfgparser['pmidcite']['dir_icite']
        dir_pubmed_txt = self.pmidcite.cfgparser.cfgparser['pmidcite']['dir_pubmed_txt']
        parser.add_argument(
            'pmids', metavar='PMID', type=int, nargs='*',
            help='PubMed IDs (PMIDs)')
        parser.add_argument(
            '-i', '--infile', nargs='*',
            help='Read PMIDs from a file containing one PMID per line.')
        parser.add_argument(
            '-a', '--append_outfile',
            help='Append current citation report to an ASCII text file. Create if needed.')
        parser.add_argument(
            '-H', '--print_header', action='store_true',
            help='Print column headings on one line.')
        parser.add_argument(
            '-k', '--print_keys', action='store_true',
            help='Print the keys describing the abbreviations.')
        parser.add_argument(
            '-o', '--outfile',
            help='Write current citation report to an ASCII text file.')
        parser.add_argument(
            '-O', action='store_true',
            help="Write each PMIDs' iCite report with citations/references to <dir_icite>/PMID.txt")
        parser.add_argument(
            '-f', '--force_write', action='store_true',
            help='if an existing outfile file exists, overwrite it.')
        parser.add_argument(
            '-D', '--force_download', action='store_true',
            help='Download PMID iCite information to a Python file, over-writing if necessary.')
        parser.add_argument(
            '--dir_icite_py', default=dir_icite_py,
            help='Write PMID iCite information into directory (default={D})'.format(D=dir_icite_py))
        parser.add_argument(
            '--dir_icite', default=dir_icite,
            help='Write PMID icite reports into directory (default={D})'.format(D=dir_icite))
        parser.add_argument(
            '-R', '--no_references', action='store_true',
            help='Print the list of citations, but not the list of references.')
        parser.add_argument(
            '-v', '--verbose', action='store_true', default=False,
            help="Print all citing papers and references for each PMID provided by user.")
        parser.add_argument(
            '-p', '--pubmed', action='store_true',
            help='Download PubMed entry containing title, abstract, authors, journal, MeSH, etc.')
        parser.add_argument(
            '--dir_pubmed_txt', default=dir_pubmed_txt,
            help='Write PubMed entry into directory (default={D})'.format(D=dir_pubmed_txt))
        self.pmidcite.cfgparser.get_nihgrouper().add_arguments(parser)
        parser.add_argument(
            '--md', action='store_true',
            help='Print using markdown table format.')
        parser.add_argument(
            '--generate-rcfile', action='store_true',
            help='Generate a sample configuration file according to the '
                 'current configuration.')
        return parser

    def cli(self):
        """Run iCite/PubMed using command-line interface"""
        argparser = self.get_argparser()
        args = argparser.parse_args()
        ## print('ICITE ARGS ../pmidcite/src/pmidcite/cli/icite.py', args)
        self.pmidcite.dir_icite_py = args.dir_icite_py
        # Print rcfile initialization file
        if args.generate_rcfile:
            prt_rcfile(stdout)
        # Print the keys and/or the header
        if args.print_keys:
            NIHiCiteDownloader.prt_keys()
        if args.print_header:
            NIHiCiteDownloader.prt_hdr()
        # Get user-specified PMIDs
        pmids = get_pmids(args.pmids, args.infile)
        if pmids:
            # Begin to initialize citation/PMID cli
            groupobj = NihGrouper(args.min1, args.min2, args.min3, args.min4)
            dnldr = self.get_icite_downloader(groupobj, args.force_download, args.no_references)
            pmid2icitepaper = dnldr.get_pmid2paper(pmids, not args.no_references, None)
            self.run_icite(pmid2icitepaper, dnldr, args, argparser)
            if args.pubmed:
                self.pubmed.dnld_wr1_per_pmid(pmids, args.force_download, args.dir_pubmed_txt)

    # pylint: disable=too-many-arguments
    def run_icite(self, pmid2icitepaper_all, dnldr, args, argparser):
        """Run iCite/PubMed"""
        pmid2icitepaper_cur = self.run_icite_pre(pmid2icitepaper_all, args, argparser)
        if not pmid2icitepaper_cur:
            return
        # Write the report with citations and references for each PMID into its own file
        if args.O:
            self._wr_papers(pmid2icitepaper_cur, dnldr)
        # Write the succinct report each PMID to the screen
        else:
            self.run_icite_wr(pmid2icitepaper_cur, args, dnldr)

    def run_icite_pre(self, pmid2icitepaper_all, args, argparser):
        """Run iCite/PubMed"""
        if not pmid2icitepaper_all and not args.print_keys and not args.print_header:
            argparser.print_help()
            self._prt_infiles(args.infile)
        pmid2icitepaper_cur = {p: o for p, o in pmid2icitepaper_all.items() if o is not None}
        if not pmid2icitepaper_cur:
            # pylint: disable=line-too-long
            self._prt_no_icite(set(pmid2icitepaper_all.keys()).difference(pmid2icitepaper_cur.keys()))
            return {}
        return pmid2icitepaper_cur

    @staticmethod
    def _prt_infiles(infiles):
        """Print input files"""
        if infiles:
            for fin in infiles:
                print('**ERROR: NO PMIDs found in: {F}'.format(F=fin))

    @staticmethod
    def run_icite_wr(pmid2icitepaper, args, dnldr):
        """Print papers, including citation counts"""
        dct = get_outfile(args.outfile, args.append_outfile, args.force_write)
        prt_verbose = args.verbose
        if dct['outfile'] is None and not args.O:
            dnldr.prt_papers(
                pmid2icitepaper, prt=stdout, prt_assc_pmids=prt_verbose)
        else:
            if args.verbose:
                dnldr.prt_papers(pmid2icitepaper, prt=stdout, prt_assc_pmids=prt_verbose)
            if dct['outfile'] is not None:
                dnldr.wr_papers(dct['outfile'], pmid2icitepaper, dct['force_write'], dct['mode'])

    def _wr_papers(self, pmid2icitepaper, dnldr):
        """Write one icite report per PMID into dir_icite/PMID.txt"""
        for pmid, paper in pmid2icitepaper.items():
            fout_txt = os.path.join(self.pmidcite.dir_icite, '{PMID}.txt'.format(PMID=pmid))
            with open(fout_txt, 'w') as prt:
                dnldr.prt_papers({pmid:paper}, prt, prt_assc_pmids=True)
                print('  WROTE: {TXT}'.format(TXT=fout_txt))

    @staticmethod
    def _prt_no_icite(pmids):
        if not pmids:
            return
        print('**NOTE: No NIH iCite papers found for: {Ps}'.format(
            Ps=' '.join(str(p) for p in pmids)))

    def get_icite_downloader(self, grouperobj, force_download, no_references):
        """Get iCite downloader"""
        # pylint: disable=line-too-long
        return self.pmidcite.get_icitedownloader(force_download, grouperobj, no_references, prt_icitepy=None)

    #def _get_pmid2icitepaper(self, pmids, grouperobj, args):
    #    """Get pmid2icitepaper"""
    #    dnldr = self.pmidcite.get_icitedownloader(args.force_download, grouperobj, args.no_references, prt=None)
    #    if args.print_keys:
    #        dnldr.prt_keys()
    #    dct = get_outfile(args.outfile, args.append_outfile, args.force_write)
    #    prt_verbose = not args.succinct
    #    return dnldr.get_pmid2paper(pmids, not args.no_references, pmid2note)


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Groups papers using the NIH percentile"""

__copyright__ = "Copyright (C) 2021-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from collections import namedtuple


class NihGrouper:
    """Groups papers using the NIH percentile"""

    ntobj = namedtuple('NtNihGroup', 'val txt')
    #### nt0 = ntobj(val=0, txt='0'),
    #### nt1 = ntobj(val=1, txt='1'),
    #### nt2 = ntobj(val=2, txt='2'),
    #### nt3 = ntobj(val=3, txt='3'),
    #### nt4 = ntobj(val=4, txt='4'),
    #### ntno = ntobj(val=5, txt='i'),

    def __init__(self, min1=2.1, min2=15.7, min3=83.9, min4=97.5):
        self.min1 = min1
        self.min2 = min2
        self.min3 = min3
        self.min4 = min4

    def get_group(self, nih_percentile):
        """Assign group numbers to the NIH percentile values using the 68-95-99.7 rule"""
        # No NIH percentile yet assigned. This paper should be checked out.
        if nih_percentile is None or nih_percentile == -1:
            return 5
        #  2.1% -3 SD: Very low citation rate
        if nih_percentile < self.min1:  # default: 2.1
            return 0
        # 13.6% -2 SD: Low citation rate
        if nih_percentile < self.min2:  # default: 15.7
            return 1
        # 68.2% -1 SD to +1 SD: Average citation rate
        if nih_percentile < self.min3:  # default: 83.9
            return 2
        # 13.6% +2 SD: High citation rate
        if nih_percentile < self.min4:  # default: 97.5
            return 3
        #  2.1% +3 SD: Very high citation rate
        return 4

    #### def get_nt(self, nih_percentile):
    ####     """Assign group numbers to the NIH percentile values using the 68-95-99.7 rule"""
    ####     # No NIH percentile yet assigned. This paper should be checked out.
    ####     if nih_percentile is None or nih_percentile == -1:
    ####         return self.ntno
    ####     #  2.1% -3 SD: Very low citation rate
    ####     if nih_percentile < self.min1:  # default: 2.1
    ####         return self.nt0
    ####     # 13.6% -2 SD: Low citation rate
    ####     if nih_percentile < self.min2:  # default: 15.7
    ####         return self.nt1
    ####     # 68.2% -1 SD to +1 SD: Average citation rate
    ####     if nih_percentile < self.min3:  # default: 83.9
    ####         return self.nt2
    ####     # 13.6% +2 SD: High citation rate
    ####     if nih_percentile < self.min4:  # default: 97.5
    ####         return self.nt3
    ####     #  2.1% +3 SD: Very high citation rate
    ####     return self.nt4

    def add_arguments(self, parser):
        """Add NIH grouper arguments to the parser"""
        # pylint: disable=line-too-long
        parser.add_argument(
            '-1', metavar='MIN1', dest='min1', default=self.min1, type=float,
            help='Minimum NIH percentile to be placed in group 1 (default: {D})'.format(D=self.min1))
        parser.add_argument(
            '-2', metavar='MIN2', dest='min2', default=self.min2, type=float,
            help='Minimum NIH percentile to be placed in group 2 (default: {D})'.format(D=self.min2))
        parser.add_argument(
            '-3', metavar='MIN3', dest='min3', default=self.min3, type=float,
            help='Minimum NIH percentile to be placed in group 3 (default: {D})'.format(D=self.min3))
        parser.add_argument(
            '-4', metavar='MIN4', dest='min4', default=self.min4, type=float,
            help='Minimum NIH percentile to be placed in group 4 (default: {D})'.format(D=self.min4))


# Copyright (C) 2021-present DV Klopfenstein. All rights reserved.



"""Given a PubMed ID (PMID), return a list of publications which cite it from NIH's iCite"""
# https://icite.od.nih.gov/api

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

from os.path import exists
from os.path import join
from sys import stdout
import collections as cx
import traceback
import requests

from pmidcite.icite.entry import NIHiCiteEntry


class NIHiCiteAPI:
    """Given a PubMed ID (PMID), return a list of publications which cite it from NIH's iCite"""

    opt_keys = {
        # Number of publications to return. The maximum allowed is 1000.
        'limit'
        # Only return publications with a PMID greater than this
        # Example: /api/pubs?offset=23456789&limit=10&format=csv
        'offset'
        # Only return publications from the given year.
        'year'
        # Only return publications with the given PubMed IDs.
        # Separate multiple IDs with commas to request up to 1000 at a time.
        # If this parameter is provided, all other parameters are ignored.
        'pmids'
        # only return publications with the given fields.
        # Separate multiple fields with commas (no space).
        # Field names are very specific and listed in Response example below.
        # No fl param will return all fields.
        # Example: /api/pubs?pmids=28968381,28324054,23843509&fl=pmid,year,title,apt
        'fl'
        # return csv (comma separated value) by specifying format=csv rather than the default JSON.
        'format'}

    url_base = 'https://icite.od.nih.gov/api/pubs'

    flds_yes_no = {'is_research_article', 'is_clinical', 'provisional'}
    yes_no = {'Yes':True, 'No':False}

    def __init__(self, nihgrouper, dirpy_dnld='.', prt=None, **kws):
        self.nihgrouper = nihgrouper
        self.dir_dnld = dirpy_dnld
        if not exists(dirpy_dnld):
            raise RuntimeError('**FATAL: NO DIRECTORY: {DIR}'.format(DIR=dirpy_dnld))
        self.prt = prt
        self.kws = {k:v for k, v in kws.items() if k in self.opt_keys}

    def dnld_icites(self, pmids):
        """Run iCite on given PubMed IDs"""
        if not pmids:
            return []
        num_pmids = len(pmids)
        if num_pmids > 1000:
            print('{N} pmids > 1000'.format(N=num_pmids))
            pmids = sorted(p for p in pmids if isinstance(p, int))
            ## print(pmids)
            print('**WARNING: USING pmids[:900]')
            pmids = pmids[:900]
        ## assert len(pmids) <= 1000, '{N} pmids > 1000'.format(N=len(pmids))
        cmd = '{URL}?pmids={PMIDS}'.format(URL=self.url_base, PMIDS=','.join(str(p) for p in pmids))
        rsp_json = self._send_request(cmd)
        lst = []
        if rsp_json is not None:
            for json_dct in rsp_json['data']:
                lst.append(self._jsonpmid_to_obj(json_dct))  # NIHiCiteEntry
        return lst

    def dnld_icite(self, pmid):
        """Run iCite on given PubMed IDs"""
        cmd = '/'.join([self.url_base, str(pmid)])
        json_dct = self._send_request(cmd)
        return self._jsonpmid_to_obj(json_dct) if json_dct else None

    def _send_request(self, cmd):
        """Send the request to iCite"""
        try:
            rsp = requests.get(cmd)
            if rsp.status_code == 200:
                return rsp.json()
            print(self._err_msg(rsp))
            return None
        except requests.exceptions.ConnectionError as errobj:
            print('**ERROR: ConnectionError = {ERR}\n'.format(ERR=str(errobj)))
            return None
        except:
            traceback.print_exc()
            raise RuntimeError('**ERROR DOWNLOADING {CMD}'.format(CMD=cmd))


    @staticmethod
    def _err_msg(rsp):
        """Get error message if an NIH iCite request failed"""
        ## print('1 RRRRRRRRRRRRRRRRRRRRRR', dir(rsp))
        ## print('2 RRRRRRRRRRRRRRRRRRRRRR', rsp.status_code)
        ## print('3 RRRRRRRRRRRRRRRRRRRRRR', rsp.reason)
        ## print('4 RRRRRRRRRRRRRRRRRRRRRR', rsp.content)
        ## print('5 RRRRRRRRRRRRRRRRRRRRRR', rsp.text)
        ## #print('3 RRRRRRRRRRRRRRRRRRRRRR', rsp.json())
        ## print('6 RRRRRRRRRRRRRRRRRRRRRR', rsp.url)
        ## if rsp.json() is not None:
        ##     txt =' '.join('{K}({V})'.format(K=k, V=v) for k, v in sorted(rsp.json().items()))
        return '{CODE} {REASON} URL[{N}]: {URL}'.format(
            CODE=rsp.status_code,
            REASON=rsp.reason,
            N=len(rsp.url),
            URL=rsp.url)
            #TEXT=rsp.text)

    def _jsonpmid_to_obj(self, json_dct):
        """Given a PMID json dict, return a NIHiCiteEntry object"""
        file_pmid = join(self.dir_dnld, 'p{PMID}.py'.format(PMID=json_dct['pmid']))
        adj_dct = self._adjust_jsondct(json_dct)
        self._wrpy(file_pmid, adj_dct)
        return NIHiCiteEntry(adj_dct, self.nihgrouper.get_group(adj_dct['nih_percentile']))

    def _adjust_jsondct(self, json_dct):
        """Adjust values in the json dict"""
        dct = {}
        if json_dct['title'] is not None:
            title = json_dct['title'].strip()
            if '"' in title:
                title = title.replace('"', "'")
            if "\n" in title:
                title = title.replace('\n', " ")
            dct['title'] = title
        if json_dct['authors'] is not None:
            dct['authors'] = json_dct['authors'].split(', ')
        yes_no = self.yes_no
        dct['is_research_article'] = yes_no[json_dct['is_research_article']]
        dct['is_clinical'] = yes_no[json_dct['is_clinical']]
        dct['provisional'] = yes_no[json_dct['provisional']]
        lst = []
        lists = {'authors', 'cited_by_clin', 'cited_by', 'references'}
        for key, val in json_dct.items():
            if key in dct:
                lst.append((key, dct[key]))
            elif key in lists:
                lst.append((key, [] if val is None else val))
            else:
                lst.append((key, val))
        return cx.OrderedDict(lst)

    def _wrpy(self, fout_py, dct):
        """Write NIH iCite to a Python module"""
        with open(fout_py, 'w') as prt:
            self.prt_dct(dct, prt)
            if self.prt:
                self.prt.write('  WROTE: {PY}\n'.format(PY=fout_py))

    def prt_dct(self, dct, prt=stdout):
        """Print NIH iCite data as a dict"""
        iciteobj = NIHiCiteEntry(dct, self.nihgrouper.get_group(dct['nih_percentile']))
        prt.write('"""Write data downloaded for NIH iCite data"""\n\n')
        prt.write('# pylint: disable=line-too-long\n')
        prt.write('# DESC: {DESC}\n'.format(DESC=str(iciteobj)))
        prt.write('ICITE = {\n')
        str_val = {'title', 'journal', 'doi'}
        for key, val in dct.items():
            if key == 'authors':
                prt.write("    '{K}': {V},\n".format(K=key, V=val))
                #prt.write("    '{K}': {AUTHORS},\n".format(
                #    K=key,
                #    AUTHORS='\n'.join(['"{AU}"'.format(AU=a) for a in val])))
            elif key not in str_val:
                prt.write("    '{K}': {V},\n".format(K=key, V=val))
            else:
                prt.write('''    '{K}': """{V}""",\n'''.format(K=key, V=val))
        prt.write('}\n')


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.



"""Run NIH iCite run for one PubMed ID (PMID)"""

__copyright__ = "Copyright (C) 2019-present, DV Klopfenstein. All rights reserved."
__author__ = "DV Klopfenstein"

import sys
from pmidcite.icite.api import NIHiCiteAPI
from pmidcite.icite.pmid_dnlder import NIHiCiteDownloader
from pmidcite.cli.utils import read_pmids


class PmidCite:
    """Run NIH iCite run for one PubMed ID (PMID)"""

    def __init__(self, cfgparser):
        self.cfgparser = cfgparser  # Cfg
        self.dir_icite = self.cfgparser.cfgparser['pmidcite']['dir_icite']  # ./log/icite
        self.dir_icite_py = self.cfgparser.cfgparser['pmidcite']['dir_icite_py']  # ./icite

    def run_pmid_file(self, fin_pmids, fout_icite, force_download):
        """Run iCite on list of PMIDs in a file"""
        pmids = read_pmids(fin_pmids)
        dnldr = self.get_icitedownloader(force_download)
        pmid2ntpaper = dnldr.get_pmid2paper(
            pmids, dnld_assc_pmids_do=False, pmid2note=None)
        dnldr.wr_papers(fout_icite, pmid2ntpaper, force_download, 'w')
        return pmids

    def prt_rcfile(self, prt=sys.stdout):
        """Print pmidcite rcfile"""
        self.cfgparser.prt_rcfile_dflt(prt)

    # pylint: disable=line-too-long
    def get_icitedownloader(self, force_download, nih_grouper=None, no_references=False, prt_icitepy=None):
        """Create NIHiCiteDownloader"""
        # Setting prt_icitepy to sys.stdout will cause: WROTE: ./icite/p10802651.py
        kws = {}  # TBD NIHiCiteCli
        if nih_grouper is None:
            nih_grouper = self.cfgparser.get_nihgrouper()
        api = NIHiCiteAPI(nih_grouper, self.dir_icite_py, prt_icitepy, **kws)
        return NIHiCiteDownloader(force_download, api, not no_references)


# Copyright (C) 2019-present DV Klopfenstein. All rights reserved.
